/***************************************************************************
                          nettransport.h  -  description
                             -------------------
    begin                : Mon Feb 25 2002
    copyright            : (C) 2002 by Dmitry Osinovsky
    email                : osdm@mail.ru
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2.1 of the    *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 ***************************************************************************/

#ifndef NETTRANSPORT_H
#define NETTRANSPORT_H

// include files for QT
#include <qobject.h>
#include <qstring.h>
#include <qsocket.h>

class IReceiver
{
public:
  virtual bool receiveOne() = 0;
  virtual void onClose() = 0;
  virtual ~IReceiver() {}
};

class QDataStream;

/** This class is an abstraction for a transport protocol.
  * @author Dmitry Osinovsky
  */

class NetTransport
{
public:
  virtual void setReceiver(IReceiver * Receiver) { FReceiver = Receiver; }
  virtual QDataStream * getIn() { return FIn; }
  virtual QDataStream * getOut() { return FOut; }
  virtual void flush() = 0;
  virtual void close() = 0;
  virtual void deleteSelf() { delete this; }
  virtual ~NetTransport() {}
protected:
  IReceiver * FReceiver;
  QDataStream * FIn;
  QDataStream * FOut;
};

/** Local protocol, no need for network connection.
  */

class LocalTransport: public virtual NetTransport
{
public:
  ~LocalTransport();
  void setPeer(LocalTransport * Peer);
  void close();
  void flush();
  void deleteSelf();
  LocalTransport(): FPeer(NULL), FFlushing(false), FDeleteMeAfterFlush(false) {}

private:
  LocalTransport * FPeer;

  // array for outgoing traffic
  QByteArray FArray;

  // Whether we are doing "flush()" function
  //   (in this case we must postpone deleting)
  bool FFlushing;

  // Whether we need to delete self after flush
  bool FDeleteMeAfterFlush;
};

class QWidget;
class ConnectionDlg;

/** TCP/IP protocol.
  */

class SocketTransport: public QSocket, public virtual NetTransport
{
  Q_OBJECT
public:
  SocketTransport(QWidget * DialogWidget);
  ~SocketTransport();
  bool connectTo(const QString & host, Q_UINT16 port);
  void flush();
  void close();
protected slots:
  void onReceive();
  void onClose();
  void onHostFound();
  void onConnected();
  void onError(int);
private:
  ConnectionDlg * FConnectionDlg;
  QWidget * FDialogWidget;
};

/** From a list builds a string with the following structure:
  * Prefix + Element1 + Delimiter + Element2 + ... + ElementN + Postfix
  * If list is empty, adds no prefix & postfix, thus returns empty string.
  */

class StringListBuilder
{
private:
  QString FPrefix, FPostfix, FDelimiter;
  QString FResult;
  bool FAddEmpty;
  int FCount;
public:
  StringListBuilder(const char * Prefix, const char * Postfix, const char * Delimiter, bool AddEmpty):
    FPrefix(Prefix), FPostfix(Postfix), FDelimiter(Delimiter),
    FResult(""), FAddEmpty(AddEmpty), FCount(0) {}

  int getCount() { return FCount; }
  void clear() { FResult = ""; FCount = 0; }

  void add(const char * Item)
  {
    if ((! FAddEmpty) && ((! Item) || (! * Item)))
      return;

    FResult += (FCount ++ ? FDelimiter : FPrefix) + Item;
  }

  QString getResult() { return FCount ? FResult + FPostfix : ""; }
};

#endif
